export MODEL="bedrock/us.anthropic.claude-3-7-sonnet-20250219-v1:0"
export AWS_ACCESS_KEY_ID="AKIAYS2NWBKCY46OWEXD"
export AWS_SECRET_ACCESS_KEY="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"
export AWS_REGION_NAME="us-east-1"